/**
 * This version uses DMA to compute very fast random numbers
 * 
 * Hunter Adams (vha3@cornell.edu)
 * converted to 320x240 with 256 colors by Bruce; brue.land@cornell.edu
 * 
 * HARDWARE CONNECTIONS
 *  - GPIO 16 ---> VGA Hsync
 *  - GPIO 17 ---> VGA Vsync
 * 
 *  - GPIO 08 ---> 330 ohm resistor ---> VGA Blue lo-bit |__ both wired to 150 ohm to ground 
 *  - GPIO 09 ---> 220 ohm resistor ---> VGA Blue hi_bit |   and to VGA blue
 * 
 *  - GPIO 10 ---> 1000 ohm resistor ---> VGA Green lo-bit |__ three wired to 100 ohm to ground
 *  - GPIO 11 ---> 680 ohm resistor ---> VGA Green mid_bit |   and to VGA Green
 *  - GPIO 12 ---> 330 ohm resistor ---> VGA Green hi_bit  |   
 * 
 *  - GPIO 13 ---> 1000 ohm resistor ---> VGA Red lo-bit |__ three wired to 100 ohm to ground
 *  - GPIO 14 ---> 680 ohm resistor ---> VGA Red mid_bit |   and to VGA red
 *  - GPIO 15 ---> 330 ohm resistor ---> VGA Red hi_bit  |   
 * 
 *  - RP2040 GND ---> VGA GND
 *
 * RESOURCES USED
 *  - PIO state machines 0 to 3 on PIO instance 0
 *  - DMA channels 0, 1, 2, 3 data send to two PIO
 *  - 76.8 kBytes of RAM (for pixel color data)
 * color encoding: bits 7:5 red; 4:2 green; 1:0 blue
 *
 * Protothreads v1.1.1
 * graphics demo thread
 * serial thread to set the distribution parameters///
 * the usual blinky thread for a hearbeat
 */
// ==========================================
// === VGA graphics library
// ==========================================
#include "vga256_graphics.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h> 
#include <math.h>
#include "pico/stdlib.h"
#include "hardware/pio.h"
#include "hardware/dma.h"

//===========================================
// === DMAcpu
//===========================================
#include "hardware/gpio.h"
#include "hardware/timer.h"
#include "hardware/spi.h"
#include "pico/stdlib.h"
#include "hardware/uart.h"
#include "hardware/gpio.h"
#include "hardware/structs/iobank0.h"
#include "hardware/regs/rosc.h"

// ==========================================
// === protothreads globals
// ==========================================
#include "hardware/sync.h"
#include "hardware/timer.h"
#include "pico/multicore.h"
#include "string.h"
#include "math.h"
// protothreads header
#include "pt_cornell_rp2040_v1_1_1.h"
// interactive color
int new_value = false ;

// ==========================================
// === DMA machine setup
// ==========================================
// === global thread + DMA communicaiton
// must be volatile because of DMA modification
// a zero load variable -- dont change
volatile int dma_zero = 0 ;
// a one-load variable -- dont change
volatile int dma_one = 1 ;
volatile int dma_neg_one = -1 ;

// noise vars
//volatile int dma_noise_temp = 0 ;
volatile int dma_rand_out = 0;
//volatile int dma_sniff_temp = 0 ;

// signal to user thead
volatile int dma_flag = 0 ;

#define sniffer_crc32 0x00
// /dev/null - a data sink
int bit_bucket ;

// ========================================
// === spi setup 
// =======================================
//SPI configurations
#define PIN_CS   5
#define PIN_SCK  6
#define PIN_MOSI 7
#define SPI_PORT spi0

// constant to tell SPI DAC what to do
// prepend to each 12-bit sample
#define DAC_config_chan_A 0b0011000000000000


// ==========================================
// === Ring Osc random bit RNG
// setup for higher speed oscillator
// ==========================================
volatile uint32_t *rnd_reg = (uint32_t *)(ROSC_BASE + ROSC_RANDOMBIT_OFFSET);
void rosc_setup(void){
  volatile uint32_t *rosc_div = (uint32_t *)(ROSC_BASE + ROSC_DIV_OFFSET) ;
  volatile uint32_t *rosc_ctl = (uint32_t *)(ROSC_BASE + ROSC_CTRL_OFFSET) ;
  volatile uint32_t *rosc_freqA = (uint32_t *)(ROSC_BASE + ROSC_FREQA_OFFSET) ;
  volatile uint32_t *rosc_freqB = (uint32_t *)(ROSC_BASE + ROSC_FREQB_OFFSET) ;
  
  // set divider to one for frequency measurement
  *rosc_div = ROSC_DIV_VALUE_PASS + 1 ;
  // speed up the ROSC so more cycles between reads
  // (dont use ROSC_CTRL_FREQ_RANGE_VALUE_TOOHIGH)
  // Measured at 241 MHz with theses settings
  *rosc_ctl =  ROSC_CTRL_FREQ_RANGE_VALUE_HIGH ;// | ROSC_CTRL_ENABLE_VALUE_ENABLE;
  *rosc_freqA = (ROSC_FREQA_PASSWD_VALUE_PASS<<16) | 0xffff ;
  *rosc_freqB = (ROSC_FREQB_PASSWD_VALUE_PASS<<16) | 0xffff ;
}

// ==========================================================
// basic random number gen with Von Neumann extractor
// and a small delay in the extractor
// AND one pass through rand() function
// https://en.wikipedia.org/wiki/Randomness_extractor
// ==========================================================

uint32_t rand_rosc_VN(void){
    int k, random ;
    int random_bit1, random_bit2 ;
    volatile uint32_t *rnd_reg = (uint32_t *)(ROSC_BASE + ROSC_RANDOMBIT_OFFSET);
    volatile uint32_t *rosc_ctl = (uint32_t *)(ROSC_BASE + ROSC_CTRL_OFFSET) ;
    
    for(k=0;k<32;k++){
      // von Neumann bit extractor         
      while(1){
        //extractor_count++ ;
        random_bit1=0x00000001 & (*rnd_reg);
        //  a small delay decorrelates the bits
        asm("nop") ; asm("nop") ;
        asm("nop") ; asm("nop") ;
        asm("nop") ; asm("nop") ;
        asm("nop") ; asm("nop") ;      
        random_bit2=0x00000001 & (*rnd_reg);
        // if the two are diferent, use the first one
        if(random_bit1!=random_bit2) break;
      }  
      // build the 32 bit sample
      random=(random << 1) | random_bit1 ;
    }
    srand(random) ;
    //rand_count++ ;
    random = rand() ;
    return random;
}



int dma_rand(int rand_mask){
    while(dma_flag==0) {} ;
    dma_flag = 0 ;
    // skip
   // while(dma_flag==0) {} ;
    //dma_flag = 0 ;
    return(dma_rand_out & rand_mask) ;
}

// ==================================================
// === hard coded 3 channel rand
// ==================================================
void DMA_rand_start(void){
// pacing timer 100 KHz set denom to 125000000/100000 = 1250
    // 200 KHz  is 625 counter divide
    // 500 KHz is 250 
    // can be user modified, or turned off if desired
    dma_timer_set_fraction ( 3, 1, 10) ;

    // enable sniffer add operation for DMA channel 1 
    // this locks sniffer to chan 1, but DMA1 has to also enable on a per-block basis
    // sniffer_add
    dma_sniffer_enable(9, sniffer_crc32, true);

    // === compute CRC
    dma_channel_config c2 = dma_channel_get_default_config(9);
    channel_config_set_transfer_data_size(&c2, DMA_SIZE_32);
    channel_config_set_read_increment(&c2, false);
    channel_config_set_write_increment(&c2, false);
    channel_config_set_irq_quiet(&c2, true);
    channel_config_set_enable(&c2, true); 
    channel_config_set_sniff_enable (&c2, true) ;
    channel_config_set_chain_to(&c2, 10) ;
    // unpaced, full speed transfer
    channel_config_set_dreq(&c2,  DREQ_FORCE); //DREQ_DMA_TIMER3
     dma_channel_configure(9, &c2, 
        &bit_bucket, // reset the fetch  to write to execute channel 
        &rnd_reg,  // read_addr, pointer to address of execute channel
        1, // one words per DMA block
        false) ; // triggered to start machine running

    // === store variable
    dma_channel_config c0 = dma_channel_get_default_config(10);
    channel_config_set_transfer_data_size(&c0, DMA_SIZE_32);
    channel_config_set_read_increment(&c0, false);
    channel_config_set_write_increment(&c0, false);
    channel_config_set_irq_quiet(&c0, true);
    channel_config_set_enable(&c0, true); 
    channel_config_set_chain_to(&c0, 11) ;
    // unpaced, full speed transfer
    channel_config_set_dreq(&c0,  DREQ_FORCE);
    //
    dma_channel_configure(10, &c0, 
        &dma_rand_out , // write to dma channel 1
        &dma_hw->sniff_data ,  // read_addr, start of DMA blocks list
        1, // four words per DMA block
        false) ; // triggered to start machine running

    // === signal main
    dma_channel_config c1 = dma_channel_get_default_config(11);
    channel_config_set_transfer_data_size(&c1, DMA_SIZE_32);
    channel_config_set_read_increment(&c1, false);
    channel_config_set_write_increment(&c1, false);
    channel_config_set_irq_quiet(&c1, true);
    channel_config_set_enable(&c1, true); 
    channel_config_set_chain_to(&c1, 9) ;
    // unpaced, full speed transfer
    channel_config_set_dreq(&c1,  DREQ_FORCE);
    //
    dma_channel_configure(11, &c1, 
        &dma_flag , // write to dma channel 1
        &dma_one ,  // read_addr, unity
        1, // four words per DMA block
        true) ; // triggered to start machine running
}
    //
// ==================================================
// === graphics demo -- RUNNING on core 0
// ==================================================
        
#define N_particle 20000
short x[N_particle] ;
short y[N_particle] ;
char  c[N_particle] ;

int start_time, end_time ;

char* array ;
#define drawPixel_m(x,y,c)  do{array[((320 * y ) + x)] = c;} while(0);
#define getPixel_m(x,y)  (array[((320*y)+x)])


static PT_THREAD (protothread_graphics(struct pt *pt)) {
    PT_BEGIN(pt);
    // the protothreads interval timer
    PT_INTERVAL_INIT() ;

    //array = (char*) getArray();

    // background
    fillRect(0, 0, 319, 239, BLACK); // 

    // Draw some filled rectangles
    fillRect(0, 0, 76, 10, BLUE); // blue box
    fillRect(100, 0, 150, 10, WHITE); // red box
    //fillRect(200, 0, 76, 10, GREEN); // green box

    // Write some text
    setTextColor(WHITE) ;
    setCursor(10, 1) ;
    setTextSize(1) ;
    writeString("ECE 4760") ;

    setTextColor(BLACK) ;
    setCursor(102, 1) ;
    setTextSize(1) ;
    writeString("VGA and DMAcpu Diffusion ") ;

    // init all the positions to random positions on screen
    for(int i=0; i<N_particle; i++){
        x[i] = (rand() % 300) + 10 ;
        y[i] = (rand() % 200) + 12 ;
        // check to see if particle landed on the center 
        if(x[i]==158 && y[i]==113) c[i]=GREEN ;
        else c[i] = WHITE ;      
    }
    // mark the center as a seed
    drawPixel(158, 113, GREEN) ;
    //drawPixel(180, 113, GREEN) ;

    //printf("%x\n\r", getPixel(160, 120)); 
    // wait for human input
    //PT_YIELD_UNTIL(pt, new_value==true) ;

    while(true) {
        start_time = PT_GET_TIME_usec() ;

        for(int i=0; i<N_particle; i++){
            // is it green?
            if (c[i]==WHITE){
              if      (getPixel(x[i]+1,y[i])==GREEN ) c[i]=GREEN ;
              else if (getPixel(x[i]-1,y[i])==GREEN ) c[i]=GREEN ;
              else if (getPixel(x[i],y[i]+1)==GREEN ) c[i]=GREEN ;
              else if (getPixel(x[i],y[i]-1)==GREEN ) c[i]=GREEN ;
            }
            //
            //drawPixel(x[i], y[i], c[i]) ;

            if(c[i]== WHITE) {
                // erase a particle
                drawPixel(x[i], y[i], BLACK) ;
                // update x,y
               // while(dma_flag==0) {} ;
               // dma_flag = 0 ;
                //x[i] += (dma_rand_out & 0b10) - 1 ;
                x[i] += dma_rand(0b10) -1 ;

                if(x[i]>318) x[i]=318 ;
                if(x[i]<1) x[i]=1 ;
                //sleep_us(1);
                // reduce serial correlation
                //while(dma_flag==0) {} ;
                //dma_flag = 0 ;
                //while(dma_flag==0) {} ;
                //dma_flag = 0 ;

                //while(dma_flag==0) {} ;
                 //       dma_flag = 0 ;
                //y[i] += (dma_rand_out & 0b10) - 1 ;
                y[i] += dma_rand(0b10) -1 ;

                if(y[i]>238) y[i]=238 ;
                if(y[i]<12) y[i]=12 ;

                //drawPixel_m(x[i], y[i], WHITE) ;

            }
            drawPixel(x[i], y[i], c[i]) ;
        }
        end_time = PT_GET_TIME_usec() ;
        setTextColor2(BLACK, WHITE) ;
        setCursor(270, 1) ;
        setTextSize(1) ;
        char vid_str[10] ;
        sprintf(vid_str, "%03d", (end_time-start_time)/1000) ;
        writeString(vid_str) ;

        //PT_YIELD_INTERVAL(20000); 
   }
   PT_END(pt);
} // graphics thread

// ==================================================
// === toggle25 thread on core 0
// ==================================================
// the on-board LED blinks
static PT_THREAD (protothread_toggle25(struct pt *pt))
{
    PT_BEGIN(pt);
    static bool LED_state = false ;
    
     // set up LED p25 to blink
     gpio_init(25) ;	
     gpio_set_dir(25, GPIO_OUT) ;
     gpio_put(25, true);
     // data structure for interval timer
     PT_INTERVAL_INIT() ;

      while(1) {
        // yield time 0.1 second
        //PT_YIELD_usec(100000) ;
        PT_YIELD_INTERVAL(100000) ;

        // toggle the LED on PICO
        LED_state = LED_state? false : true ;
        gpio_put(25, LED_state);

        // sanity check print
        //printf("%8x\n\r", dma_rand_out) ;
        //
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // blink thread


// ==================================================
// === user's serial input thread on core 0
// ==================================================
// serial_read an serial_write do not block any thread
// except this one
static int r=7, g=7, b=3 ;
float h,s,v;
static PT_THREAD (protothread_serial(struct pt *pt))
{
    PT_BEGIN(pt);
        char video_buffer[20];
      //
      while(1) {
        // print prompt
        sprintf(pt_serial_out_buffer, "1 to start: ");
        // spawn a thread to do the non-blocking write
        serial_write ;

        // spawn a thread to do the non-blocking serial read
         serial_read ;
        sscanf(pt_serial_in_buffer, "%d  ", &new_value) ;
        //printf("%d\n\r" bit_bucket) ;
        // 
        new_value = true ;
        

        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // serial thread

// ========================================
// === core 1 main -- started in main below
// ========================================
void core1_main(){ 
  //
  //  === add threads  ====================
  // for core 1
  //pt_add_thread(protothread_toggle_gpio4) ;
  //pt_add_thread(protothread_serial) ;
  //
  // === initalize the scheduler ==========
  pt_schedule_start ;
  // NEVER exits
  // ======================================
}

// ========================================
// === core 0 main
// ========================================
int main(){
  // set the clock
  //set_sys_clock_khz(250000, true); // 171us
  // start the serial i/o
  stdio_init_all() ;
  // announce the threader version on system reset
  printf("\n\rProtothreads RP2040 v1.11 two-core\n\r");

  // Initialize the VGA screen
  initVGA() ;

  // Initialize SPI channel (channel, baud rate set to 20MHz)
    // connected to spi DAC
    spi_init(SPI_PORT, 20000000) ;
    // Format (channel, data bits per transfer, polarity, phase, order)
    spi_set_format(SPI_PORT, 16, 0, 0, 0);
    // Map SPI signals to GPIO ports
    //gpio_set_function(PIN_MISO, GPIO_FUNC_SPI);
    gpio_set_function(PIN_SCK, GPIO_FUNC_SPI);
    gpio_set_function(PIN_MOSI, GPIO_FUNC_SPI);
    gpio_set_function(PIN_CS, GPIO_FUNC_SPI) ;

    //start the ROSC and speed it up
    rosc_setup();

    // seed the CRC computation
    (dma_hw->sniff_data) = rand_rosc_VN() ;

    DMA_rand_start();

    // define the DMA program
    //DMA_machine_program() ;

    // turn on the DMA coprocessor   
    // runs automomously after this
    //DMA_machine_start() ;
     
  // start core 1 threads
  //multicore_reset_core1();
  //multicore_launch_core1(&core1_main);

  // === config threads ========================
  // for core 0
  pt_add_thread(protothread_graphics);
  pt_add_thread(protothread_toggle25);
  pt_add_thread(protothread_serial) ;
  //
  // === initalize the scheduler ===============
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main